from django.urls import path

from drf_payments import get_payment_service
from drf_payments.braintree import BraintreeProvider
from drf_payments.mixins import PaymentCallbackView, PaymentSettingsView

urlpatterns = [
    path("callback", PaymentCallbackView.as_view(), name="payment-callback"),
]
# TODO: Check if other payments need settings view
if isinstance(get_payment_service("braintree"), BraintreeProvider):
    urlpatterns.append(path("settings", PaymentSettingsView().as_view(), name="braintree-settings"))



from dataclasses import asdict, dataclass, field
from typing import Optional

import stripe

from ..constants import PaymentError, PaymentStatus
from ..core import BasicProvider


def convert_amount(currency, amount) -> int:
    """convert_amount

    Converts amount from decimal to integer cents

    Args:
        currency (currency): Your currency code
        amount (decimal): decimal amount

    Returns:
        _type_: _description_
    """
    factor = 100 if currency.lower() not in zero_decimal_currency else 1
    return int(amount * factor)


@dataclass
class StripeProductData:
    name: str
    description: Optional[str] = field(init=False, repr=False, default=None)
    images: Optional[str] = field(init=False, repr=False, default=None)
    metadata: Optional[dict] = field(init=False, repr=False, default=None)
    tax_code: Optional[str] = field(init=False, repr=False, default=None)


@dataclass
class StripePriceData:
    currency: str
    product_data: StripeProductData
    unit_amount: int
    recurring: Optional[dict] = field(init=False, repr=False, default=None)
    tax_behavior: Optional[str] = field(init=False, repr=False, default=None)


@dataclass
class StripeLineItem:
    price_data: StripePriceData
    quantity: int
    adjustable_quantity: Optional[dict] = field(init=False, repr=False, default=None)
    dynamic_tax_rates: Optional[dict] = field(init=False, repr=False, default=None)
    tax_rates: Optional[str] = field(init=False, repr=False, default=None)


zero_decimal_currency = [
    "bif",
    "clp",
    "djf",
    "gnf",
    "jpy",
    "kmf",
    "krw",
    "mga",
    "pyg",
    "rwf",
    "ugx",
    "vnd",
    "vuv",
    "xaf",
    "xof",
    "xpf",
]


class StripeCheckoutProvider(BasicProvider):
    """StripeCheckoutProvider

    StripeCheckoutProvider for Stripe Checkout payments
    it handles:

    - Creating a Checkout Session

    - Refunding payment

    - Process payment confirmation with callback

    Args:
        secret_key (string): Your stripe secret_key
    """

    def __init__(self, secret_key, **kwargs):
        super().__init__(**kwargs)
        self.secret_key = secret_key

    def process_payment(self, payment):
        """process_payment

        Process payment instance via StripeCheckoutProvider

        Args:
            payment (payment): Your app payment instance

        """
        if payment.transaction_id:
            raise PaymentError("This payment has already been processed.")
        stripe.api_key = self.secret_key
        session_data = {
            "line_items": self.get_line_items(payment),
            "mode": "payment",
            "success_url": payment.success_url,
            "cancel_url": payment.failure_url,
            "client_reference_id": payment.pk,
        }
        # Patch session with billing email if exists
        if payment.billing_email:
            session_data["customer_email"] = payment.billing_email
        try:
            session = stripe.checkout.Session.create(**session_data)
            payment.transaction_id = session.get("id", None)
            payment.extra_data["session"] = session
            payment.save(update_fields=["extra_data", "transaction_id"])
            return session

        except stripe.error.StripeError as e:
            raise PaymentError(e) from e

    def refund(self, payment, amount=None):
        """refund

        Refund payment instance

        Args:
            payment (payment): Your payment instance
            amount (int, optional): Amount to refund. Defaults to None.

        """
        if payment.status == PaymentStatus.CONFIRMED.name:
            to_refund = amount or payment.total
            payment_intent = payment.extra_data.get("session", {}).get("payment_intent", None)
            if not payment_intent:
                raise PaymentError("Can't Refund, payment_intent does not exist")
            stripe.api_key = self.secret_key
            try:
                refund = stripe.Refund.create(
                    payment_intent=payment_intent,
                    amount=convert_amount(payment.currency, to_refund),
                    reason="requested_by_customer",
                )
            except stripe.error.StripeError as e:
                raise PaymentError(e) from e
            else:
                payment.extra_data["refund"] = refund
                payment.status = PaymentStatus.REFUNDED.name
                payment.save(update_fields=["extra_data", "status"])

                return convert_amount(payment.currency, to_refund)

        raise PaymentError("Only Confirmed payments can be refunded")

    def get_line_items(self, payment):
        """get_line_items

        Construct `line_items` from payment

        Args:
            payment (payment): Your payment instance

        """
        order_no = payment.pk
        product_data = StripeProductData(name=f"Order #{order_no}")

        price_data = StripePriceData(
            currency=payment.currency.lower(),
            unit_amount=convert_amount(payment.currency, payment.total),
            product_data=product_data,
        )
        line_item = StripeLineItem(
            quantity=1,
            price_data=price_data,
        )
        return [asdict(line_item)]


class StripeProvider(BasicProvider):
    """StripeProvider
    Creating payment based on `payment_method` created on FE part

    it handles:

    - Creating a payment intent

    - Refunding payment

    - Receiving payment confirmation with callback

    Args:
        secret_key (string): Your stripe secret_key
        public_key (string): Your stripe public_key
    """

    def __init__(self, secret_key, public_key, **kwargs):
        super().__init__(**kwargs)
        self.secret_key = secret_key
        self.public_key = public_key

    def process_payment(self, payment):
        """process_payment

        Process payment, require `payment_method.id` to be passed in payment.transaction_id
        upon creation

        Args:
            payment (payment): Payment instance
        """
        stripe.api_key = self.secret_key
        # * Create payment intent with payment method generated on FE
        intent_data = {
            "payment_method": payment.transaction_id,
            "amount": int(payment.total * 100),
            "currency": payment.currency,
            "confirmation_method": "automatic",
            "confirm": True,
            "metadata": {"order_no": payment.pk},
        }
        try:
            payment_intent = stripe.PaymentIntent.create(**intent_data)
        except stripe.error.StripeError as e:
            raise PaymentError(e) from e
        payment.extra_data["payment_intent"] = payment_intent
        # * Switching transaction id to payment intent_id
        payment.transaction_id = payment_intent.get("id", None)
        payment.save(update_fields=["extra_data", "transaction_id"])

    def refund(self, payment, amount=None):
        """refund

        Refund payment instance

        Args:
            payment (payment): Your payment instance
            amount (int, optional): Amount to refund. Defaults to None.

        """
        if payment.status == PaymentStatus.CONFIRMED.name:
            to_refund = amount or payment.total
            payment_intent = payment.extra_data.get("payment_intent", None).get("id", None)
            if not payment_intent:
                raise PaymentError("Can't Refund, payment_intent does not exist")
            stripe.api_key = self.secret_key
            try:
                refund = stripe.Refund.create(
                    payment_intent=payment_intent,
                    amount=convert_amount(payment.currency, to_refund),
                    reason="requested_by_customer",
                )
            except stripe.error.StripeError as e:
                raise PaymentError(e) from e
            else:
                payment.extra_data["refund"] = refund
                payment.status = PaymentStatus.REFUNDED.name
                payment.save(update_fields=["extra_data", "status"])
                return convert_amount(payment.currency, to_refund)

        raise PaymentError("Only Confirmed payments can be refunded")



from enum import Enum


class PaymentError(Exception):
    def __init__(self, message, code=None, gateway_message=None):
        super().__init__(message)
        self.code = code
        self.gateway_message = gateway_message


class PaymentStatus(Enum):
    WAITING = "waiting"
    PREAUTH = "preauth"
    CONFIRMED = "confirmed"
    REJECTED = "rejected"
    REFUNDED = "refunded"
    ERROR = "error"
    INPUT = "input"


class FraudStatus(Enum):
    UNKNOWN = "unknown"
    ACCEPT = "accept"
    REJECT = "reject"
    REVIEW = "review"


class PaymentCurrency(Enum):
    CAD = "cad"
    USD = "usd"
    EUR = "eur"
    GBP = "gbp"
    AUD = "aud"
    JPY = "jpy"
    CHF = "chf"
    HKD = "hkd"
    NZD = "nzd"
    SGD = "sgd"
    SEK = "sek"
    DKK = "dkk"
    NOK = "nok"
    MXN = "mxn"
    BRL = "brl"
    MYR = "myr"
    PHP = "php"
    THB = "thb"
    IDR = "idr"
    TRY = "try"
    INR = "inr"
    RUB = "rub"
    ILS = "ils"
    SAR = "sar"



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField

from .constants import FraudStatus, PaymentCurrency, PaymentStatus


class BasePayment(models.Model):
    """
    Model to represent single payment transaction
    """

    variant = models.CharField(max_length=255)
    #: Transaction status
    status = models.CharField(
        max_length=255,
        choices=[(v.name, v.value) for v in PaymentStatus],
        blank=True,
        default=PaymentStatus.WAITING.name,
    )
    fraud_status = models.CharField(
        _("fraud check"),
        max_length=10,
        choices=[(v.name, v.value) for v in FraudStatus],
        default=FraudStatus.UNKNOWN.name,
    )
    fraud_message = models.TextField(blank=True, default="")
    #: Creation date and time
    created = models.DateTimeField(auto_now_add=True)
    #: Date and time of last modification
    modified = models.DateTimeField(auto_now=True)
    #: Transaction ID (if applicable)
    transaction_id = models.CharField(max_length=255, blank=True)
    #: Currency code (may be provider-specific)
    currency = models.CharField(
        _("Currency"),
        max_length=10,
        choices=[(v.name, v.value) for v in PaymentCurrency],
        default=PaymentCurrency.USD.name,
    )
    #: Total amount (gross)
    total = models.DecimalField(max_digits=9, decimal_places=2, default=0.00)
    delivery = models.DecimalField(max_digits=9, decimal_places=2, default=0.00)
    tax = models.DecimalField(max_digits=9, decimal_places=2, default=0.00)
    description = models.TextField(blank=True, default="")
    billing_first_name = models.CharField(max_length=256, blank=True)
    billing_last_name = models.CharField(max_length=256, blank=True)
    billing_address_1 = models.CharField(max_length=256, blank=True)
    billing_address_2 = models.CharField(max_length=256, blank=True)
    billing_city = models.CharField(max_length=256, blank=True)
    billing_postcode = models.CharField(max_length=256, blank=True)
    billing_country_code = models.CharField(max_length=2, blank=True)
    billing_country_area = models.CharField(max_length=256, blank=True)
    billing_email = models.EmailField(blank=True)
    billing_phone = PhoneNumberField(blank=True)
    customer_ip_address = models.GenericIPAddressField(blank=True, null=True)
    extra_data = models.JSONField(default=dict)
    message = models.TextField(blank=True, default="")
    token = models.CharField(max_length=36, blank=True, default="")
    captured_amount = models.DecimalField(max_digits=9, decimal_places=2, default=0.00)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.variant}-{self.total}"

    @property
    def failure_url(self) -> str:
        return f"{settings.PAYMENT_FAILURE_URL}"

    @property
    def success_url(self) -> str:
        return f"{settings.PAYMENT_SUCCESS_URL}"



import importlib
from decimal import Decimal
from typing import NamedTuple, Optional, Union

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from drf_payments.authorizenet import AuthorizeNetProvider
from drf_payments.braintree import BraintreeProvider
from drf_payments.paypal import PaypalProvider
from drf_payments.stripe import StripeCheckoutProvider, StripeProvider


class PurchasedItem(NamedTuple):
    """A single item in a purchase."""

    name: str
    quantity: int
    price: Decimal
    currency: str
    sku: str
    tax_rate: Optional[Decimal] = None


def get_payment_model():
    """
    Method to get payment model from project settings
    """
    try:
        app_label, model_name = settings.PAYMENT_MODEL.split(".")
    except (ValueError, AttributeError) as e:
        raise ImproperlyConfigured("PAYMENT_MODEL must be of the form " '"app_label.model_name"') from e
    payment_model = apps.get_model(app_label, model_name)
    if payment_model is None:
        msg = f'PAYMENT_MODEL refers to model "{settings.PAYMENT_MODEL}" that has not been installed'
        raise ImproperlyConfigured(msg)
    return payment_model


def get_payment_service(
    variant=None,
) -> Union[BraintreeProvider, StripeCheckoutProvider, StripeProvider, PaypalProvider, AuthorizeNetProvider]:
    """Returns instance of payment service based on variant"""
    if variant == "stripe":
        module, service_name = settings.PAYMENT_VARIANTS.get(variant)[0].rsplit(".", 1)
        return getattr(importlib.import_module(module), service_name)(
            secret_key=settings.PAYMENT_VARIANTS.get(variant)[1]["secret_key"],
            public_key=settings.PAYMENT_VARIANTS.get(variant)[1]["public_key"],
        )
    elif variant == "paypal":
        module, service_name = settings.PAYMENT_VARIANTS.get(variant)[0].rsplit(".", 1)
        return getattr(importlib.import_module(module), service_name)(
            client_id=settings.PAYMENT_VARIANTS.get(variant)[1]["client_id"],
            secret_key=settings.PAYMENT_VARIANTS.get(variant)[1]["secret"],
            endpoint=settings.PAYMENT_VARIANTS.get(variant)[1]["endpoint"],
        )
    elif variant == "braintree":
        module, service_name = settings.PAYMENT_VARIANTS.get(variant)[0].rsplit(".", 1)
        return getattr(importlib.import_module(module), service_name)(
            merchant_id=settings.PAYMENT_VARIANTS.get(variant)[1]["merchant_id"],
            public_key=settings.PAYMENT_VARIANTS.get(variant)[1]["public_key"],
            private_key=settings.PAYMENT_VARIANTS.get(variant)[1]["private_key"],
            sandbox=settings.PAYMENT_VARIANTS.get(variant)[1]["sandbox"],
        )
    elif variant == "authorizenet":
        module, service_name = settings.PAYMENT_VARIANTS.get(variant)[0].rsplit(".", 1)
        return getattr(importlib.import_module(module), service_name)(
            login_id=settings.PAYMENT_VARIANTS.get(variant)[1]["login_id"],
            transaction_key=settings.PAYMENT_VARIANTS.get(variant)[1]["transaction_key"],
            endpoint=settings.PAYMENT_VARIANTS.get(variant)[1]["endpoint"],
        )
    else:
        raise ImproperlyConfigured(f"{variant} is not valid variant")



from django.shortcuts import get_object_or_404
from rest_framework import generics, serializers, views
from rest_framework.decorators import action
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

import drf_payments
from drf_payments import get_payment_model, get_payment_service
from drf_payments.constants import PaymentStatus


class PaymentSerializerMixin(serializers.ModelSerializer):
    """PaymentSerializerMixin

    Should be used on your payment model, will handle payment related flow on creation of Payment Instance

    """

    card = serializers.CharField(required=False, write_only=True)
    card_expiration = serializers.CharField(required=False, write_only=True)
    card_cvv = serializers.CharField(required=False, write_only=True)

    class Meta:
        model = get_payment_model()
        fields = "__all__"
        read_only_fields = ["status", "extra_data"]

    def validate(self, attrs):
        # * In case of authorizenet provider we must provide card data
        if self.initial_data.get("variant") == "authorizenet" and all(
            key not in attrs for key in ["card", "card_cvv", "card_expiration"]
        ):
            raise serializers.ValidationError("Card, card_expiration, card_cvv are required when using authorizenet")
        return super().validate(attrs)

    def create(self, validated_data):
        # * Move card data to extra data field in case of authorizenet provider
        if validated_data["variant"] == "authorizenet":
            validated_data["extra_data"] = {
                "card": {
                    "x_card_num": validated_data.pop("card"),
                    "x_card_code": validated_data.pop("card_cvv"),
                    "x_exp_date": validated_data.pop("card_expiration"),
                },
            }
        instance = super().create(validated_data)
        service = get_payment_service(instance.variant)
        service.process_payment(instance)
        return instance

    # ? Adding payment url from extra_data
    def to_representation(self, instance):
        """
        Override the default representation of the instance object to include the payment urls
        """
        data = super().to_representation(instance)
        if instance.variant == "stripe" and isinstance(
            get_payment_service(instance.variant),
            drf_payments.stripe.StripeCheckoutProvider,
        ):
            data["url"] = instance.extra_data["session"]["url"]
        # * In case of paypal we return url for checkout form
        if instance.variant == "paypal" and isinstance(
            get_payment_service(instance.variant),
            drf_payments.paypal.PaypalProvider,
        ):
            data["url"] = instance.extra_data["order"]["links"][1]["href"]
        return data


class PaymentViewMixin(ModelViewSet):
    "Add custom method for payment instance based on variant"
    serializer_class = PaymentSerializerMixin
    queryset = get_payment_model().objects.all()

    @action(detail=True, methods=["POST"])
    def refund(self, request, pk):
        payment = get_object_or_404(get_payment_model())
        try:
            get_payment_service(payment.variant).refund(payment)
        except Exception as e:
            return Response(data={"error": str(e)}, status=400)
        return Response(data=self.serializer_class(payment).data, status=200)


class PaymentCallbackSerializerMixin(serializers.Serializer):
    """
    StripeCheckoutProvider

    success stripe checkout session send 4 events:

    payment_intent.created

    charge.succeeded

    payment_intent.succeeded

    checkout.session.completed

    We will capture checkout.session.completed and modify our payment model data.
    example data:
    ```
    StripeProvider

    Send 3 events:

    payment_intent.created

    charge.succeeded

    payment_intent.succeeded

    PayPalProvider

    Send event

    CHECKOUT.ORDER.APPROVED


    """

    def create(self, validated_data):
        event = self.context.get("request", None).data
        # * Stripe checkout session hook
        if event.get("type") == "checkout.session.completed":
            payment_id = event["data"]["object"].get("client_reference_id")
            status = event["data"]["object"]["payment_status"]
            if status == "paid":
                try:
                    self._change_payment_status(payment_id, event, "session")
                except Exception as e:
                    raise serializers.ValidationError(f"Payment with id {payment_id} not found") from e
        # * Stripe payment  hook
        if event.get("type") == "payment_intent.succeeded":
            payment_id = event["data"]["object"].get("metadata", {}).get("order_no", None)
            status = event["data"]["object"]["status"]
            if status == "succeeded":
                try:
                    self._change_payment_status(payment_id, event, "payment_intend")
                except Exception as e:
                    raise serializers.ValidationError(f"Payment with id {payment_id} not found") from e
        # * Paypal hook
        if event.get("event_type") == "CHECKOUT.ORDER.APPROVED":
            payment_id = event.get("resource", {}).get("id")
            status = event.get("resource", {}).get("status")
            # * Upon checkout approval we change status and capture payment
            if status == "APPROVED":
                try:
                    payment = get_payment_model().objects.get(transaction_id=payment_id)
                    payment.status = PaymentStatus.CONFIRMED.name
                    payment.extra_data["order"] = event.get("resource", {})
                    payment.save(update_fields=["status", "extra_data"])
                    get_payment_service("paypal").capture(payment)
                except Exception as e:
                    raise serializers.ValidationError(f"Payment with id {payment_id} not found") from e
        # * Braintree webhook
        if "bt_signature" in event:
            bt = get_payment_service("braintree")
            try:
                result = bt.service.webhook_notification.parse(event["bt_signature"], event["bt_payload"])
            except Exception as e:
                raise serializers.ValidationError(f"Can't parse event {e}") from e
            try:
                payment = get_payment_model().objects.get(transaction_id=result.transaction.id)
                data = bt._serialize(bt.service.transaction.find(result.transaction.id).__dict__)
                payment.extra_data["transaction"] = data
                payment.status = PaymentStatus.CONFIRMED.name
                payment.save(update_fields=["status", "extra_data"])
            except Exception as e:
                raise serializers.ValidationError(f"Can't find payment {result.transaction.id}") from e
        return validated_data

    @staticmethod
    def _change_payment_status(payment_id, event, data):
        payment = get_payment_model().objects.get(pk=payment_id)
        payment.status = PaymentStatus.CONFIRMED.name
        payment.extra_data[data] = event["data"]["object"]
        payment.save(update_fields=["status", "extra_data"])

    def to_representation(self, instance):
        return {"message": "Your payment was successful"}


class PaymentCallbackView(generics.CreateAPIView):
    """Override your serializer_class"""

    serializer_class = PaymentCallbackSerializerMixin
    permission_classes = (AllowAny,)


class PaymentSettingsView(views.APIView):
    permission_classes = (AllowAny,)

    def get(self, request, *args, **kwargs):
        try:
            token = get_payment_service("braintree").get_client_token()
        except Exception:
            return Response(status=400)

        return Response(status=200, data={"client_token": token})



import base64

import requests
from django.conf import settings

from drf_payments.constants import PaymentError, PaymentStatus
from drf_payments.core import BasicProvider


class PaypalProvider(BasicProvider):
    """PaypalProvider


    it handles:

    - Creating a Checkout Session

    - Refunding payment

    - Process payment confirmation with callback

    Args:
        client_id (string): Your paypal client_id
        secret_key (string): Your paypal secret_key
        endpoint (url): Paypal endpoint sanbox or production
    """

    def __init__(self, client_id, secret_key, endpoint, **kwargs):
        super().__init__(**kwargs)
        self.client_id = client_id
        self.secret_key = secret_key
        self.endpoint = endpoint

    def process_payment(self, payment):
        """process_payment

        Process payment instance via PaypalProvider

        Args:
            payment (payment): Your app payment instance

        """
        if payment.transaction_id:
            raise PaymentError("This payment has already been processed.")
        token = self._create_token()
        payload = {
            "intent": "CAPTURE",
            "application_context": {
                "return_url": settings.PAYMENT_SUCCESS_URL,
                "cancel_url": settings.PAYMENT_FAILURE_URL,
            },
            "purchase_units": [
                {
                    "amount": {
                        "currency_code": payment.currency,
                        "value": float(payment.total),
                    },
                },
            ],
        }
        try:
            resp = requests.post(
                f"{self.endpoint}/v2/checkout/orders",
                headers={"Authorization": f"Bearer {token}"},
                json=payload,
            ).json()
        except requests.exceptions.RequestException as e:
            raise PaymentError(e) from e
        payment.transaction_id = resp.get("id")
        payment.extra_data["order"] = resp
        payment.save(update_fields=["extra_data", "transaction_id"])

    def _create_token(self) -> str:
        """_create_token

        Method for creating authorization token for PayPal requests

        Returns:
            str: _description_
        """
        token = base64.b64encode(f"{self.client_id}:{self.secret_key}".encode("utf-8")).decode("utf-8")

        if access_token := (
            requests.post(
                f"{self.endpoint}/v1/oauth2/token",
                data={"grant_type": "client_credentials"},
                headers={"Authorization": f"Basic {token}"},
            )
            .json()
            .get("access_token")
        ):
            return access_token
        else:
            raise PaymentError("Can't create token")

    def refund(self, payment, amount=None):
        """refund

        Refund payment instance

        Args:
            payment (payment): Your payment instance
            amount (int, optional): Amount to refund. Defaults to None.

        """
        if payment.status == PaymentStatus.CONFIRMED.name:
            try:
                capture = payment.extra_data["order"]["purchase_units"][0]["payments"]["captures"][0]["id"]
            except (IndexError, KeyError) as e:
                raise PaymentError("Can't Refund, payment has not been captured yet") from e
            token = self._create_token()
            resp = requests.post(
                f"{self.endpoint}/v2/payments/captures/{capture}/refund",
                headers={"Authorization": f"Bearer {token}"},
                json={},
            ).json()
            payment.extra_data["order"] = resp
            payment.save(update_fields=["extra_data"])
            return
        raise PaymentError("Only Confirmed payments can be refunded")

    def capture(self, payment):
        """capture

        Upon successful checkout we will capture funds to finalize payment and have ability to refund

        Args:
            payment (payment): Your payment
        """
        token = self._create_token()
        resp = requests.post(
            f"{self.endpoint}/v2/checkout/orders/{payment.transaction_id}/capture",
            headers={"Authorization": f"Bearer {token}"},
            json={},
        ).json()
        payment.extra_data["order"] = resp
        payment.save(update_fields=["extra_data"])



import braintree

from drf_payments.constants import PaymentError, PaymentStatus
from drf_payments.core import BasicProvider


class BraintreeProvider(BasicProvider):
    def __init__(self, merchant_id, public_key, private_key, sandbox, **kwargs):
        super().__init__(**kwargs)

        if sandbox:
            self.service = braintree.BraintreeGateway(
                braintree.Configuration(
                    braintree.Environment.Sandbox,
                    merchant_id=merchant_id,
                    public_key=public_key,
                    private_key=private_key,
                ),
            )
        else:
            self.service = braintree.BraintreeGateway(
                braintree.Configuration(
                    braintree.Environment.Production,
                    merchant_id=merchant_id,
                    public_key=public_key,
                    private_key=private_key,
                ),
            )

    def process_payment(self, payment):
        payment_method = payment.transaction_id
        try:
            result = self.service.transaction.sale(
                {
                    "amount": str(float(payment.total)),
                    "payment_method_nonce": payment_method,
                    "options": {"submit_for_settlement": True},
                },
            )
        except Exception as e:
            raise PaymentError("Can't process payment") from e

        data = self._serialize(result.transaction.__dict__)
        payment.transaction_id = result.transaction.id
        payment.extra_data["transaction"] = data
        payment.save(update_fields=["extra_data", "transaction_id"])

    def refund(self, payment, amount=None):
        if payment.status == PaymentStatus.CONFIRMED.name:
            try:
                result = self.service.transaction.refund(payment.transaction_id)
                if isinstance(result, braintree.ErrorResult):
                    raise PaymentError(f"Can't process refund: {result.message}")
                payment.status = PaymentStatus.REFUNDED.name
                payment.save(update_fields=["status"])
                return
            except Exception as e:
                raise PaymentError("Can't process refund") from e
        raise PaymentError("Only Confirmed payments can be refunded")

    @staticmethod
    def _serialize(data) -> dict:
        """Helper method to serialize transaction data"""
        skip_fields = [
            "_setattrs",
            "disbursement_details",
            "descriptor",
            "status_history",
            "gateway",
            "billing_details",
            "credit_card_details",
            "shipping_details",
            "subscription_details",
            "customer_details",
        ]
        result = {key: item for key, item in data.items() if key not in skip_fields}
        result["amount"] = str(result["amount"])
        result["created_at"] = str(result["created_at"])
        result["updated_at"] = str(result["updated_at"])
        result["authorization_expires_at"] = str(result["authorization_expires_at"])
        return result

    def get_client_token(self):
        try:
            token = self.service.client_token.generate()
        except Exception:
            return None
        return token



import requests

from drf_payments.constants import PaymentStatus

from ..core import BasicProvider

RESPONSE_STATUS = {
    "1": PaymentStatus.CONFIRMED,
    "2": PaymentStatus.REJECTED,
}


class AuthorizeNetProvider(BasicProvider):
    """Payment provider for Authorize.Net.

    This backend implements payments using the Advanced Integration Method (AIM) from
    `Authorize.Net <https://www.authorize.net/>`_.
    Due to test environment not create actual transaction currently can't implement refund method
    """

    def __init__(self, login_id, transaction_key, endpoint="https://test.authorize.net/gateway/transact.dll", **kwargs):
        self.login_id = login_id
        self.transaction_key = transaction_key
        self.endpoint = endpoint

    def process_payment(self, payment):
        """Due to official SDK don't support python 3.10 onwards  manual implementation required"""
        data = {
            "x_amount": payment.total,
            "x_refId": payment.id,
            "x_currency_code": payment.currency,
            "x_description": payment.description,
            "x_first_name": payment.billing_first_name,
            "x_last_name": payment.billing_last_name,
            "x_address": f"{payment.billing_address_1}, {payment.billing_address_2}",
            "x_city": payment.billing_city,
            "x_zip": payment.billing_postcode,
            "x_country": payment.billing_country_area,
            "x_customer_ip": payment.customer_ip_address,
            "x_login": self.login_id,
            "x_tran_key": self.transaction_key,
            "x_delim_data": True,
            "x_delim_char": "|",
            "x_method": "CC",
            "x_type": "AUTH_CAPTURE",
        }
        # *  Append card data to payload
        data.update(payment.extra_data["card"])
        resp = requests.post(self.endpoint, data=data)
        data = resp.text.split("|")
        message = data[3]
        status = "error"
        if resp.ok and RESPONSE_STATUS.get(data[0], False):
            status = RESPONSE_STATUS.get(data[0], status)
            payment.transaction_id = data[6]
            payment.status = status.name
            payment.save(update_fields=["transaction_id", "status"])
        else:
            payment.status = PaymentStatus.ERROR.name
            payment.extra_data["errors"] = [message]
            payment.save(update_fields=["status", "extra_data"])



from typing import Dict, Tuple

from django.conf import settings
from django.utils.module_loading import import_string

PAYMENT_VARIANTS: Dict[str, Tuple[str, Dict]] = {"default": ("payments.dummy.DummyProvider", {})}


class BasicProvider:
    """Defined a base provider API.

    All providers backends should subclass this class.

    ``BasicProvider`` should not be instantiated directly. Use factory instead.
    """

    def __init__(self, capture=True, **kwargs):
        """Create a new provider instance.

        This method should not be called directly; use :func:`provider_factory`
        instead.
        """
        self._capture = capture

    def process_payment(self, payment):
        raise NotImplementedError()

    def refund(self, payment, amount=None):
        raise NotImplementedError()


PROVIDER_CACHE = {}


def _default_provider_factory(variant: str, payment=None):
    """Return the provider instance based on ``variant``.

    :arg variant: The name of a variant defined in ``PAYMENT_VARIANTS``.
    """
    variants = getattr(settings, "PAYMENT_VARIANTS", PAYMENT_VARIANTS)
    handler, config = variants.get(variant, (None, None))
    if not handler:
        raise ValueError(f"Payment variant does not exist: {variant}")
    if variant not in PROVIDER_CACHE:  # pragma no branch
        module_path, class_name = handler.rsplit(".", 1)
        module = __import__(str(module_path), globals(), locals(), [str(class_name)])
        class_ = getattr(module, class_name)
        PROVIDER_CACHE[variant] = class_(**config)
    return PROVIDER_CACHE[variant]


if PAYMENT_VARIANT_FACTORY := getattr(settings, "PAYMENT_VARIANT_FACTORY", None):
    provider_factory = import_string(PAYMENT_VARIANT_FACTORY)
else:
    provider_factory = _default_provider_factory
